;; Generated package description from num3-mode.el  -*- no-byte-compile: t -*-
(define-package "num3-mode" "1.5" "highlight groups of digits in long numbers" 'nil :commit "fdf4d633e541875a6f765f5c3c8e0d4911122ed3" :url "https://elpa.gnu.org/packages/num3-mode.html" :authors '(("Felix Lee" . "felix8a@gmail.com") ("Michal Nazarewicz" . "mina86@mina86.com")) :maintainer '("Michal Nazarewicz" . "mina86@mina86.com") :keywords '("faces" "minor-mode"))
